LATENT_HEAT_OF_VAPORIZATION = 2.5e6  # J/kg
LATENT_HEAT_OF_FREEZING = 334000.0  # J/kg
GRAVITY = 9.80665  # m/s^2
# following values are used by SHiELD's slab ocean model, and so we follow suit here.
SPECIFIC_HEAT_OF_WATER = 4000.0  # J/kg/K
DENSITY_OF_WATER = 1000.0  # kg/m^3

SPECIFIC_HEAT_OF_DRY_AIR_CONST_PRESSURE = 1004.6  # J/kg/K

RVGAS = 461.5  # J/kg/K
RDGAS = 287.05  # J/kg/K

SPECIFIC_HEAT_OF_DRY_AIR_CONST_VOLUME = SPECIFIC_HEAT_OF_DRY_AIR_CONST_PRESSURE - RDGAS

SPECIFIC_HEAT_OF_WATER_CM4 = 3992.0  # J/kg/K
DENSITY_OF_WATER_CM4 = 1035.0  # kg/m^3
